@@ -119,8 +119,14 @@ match_host(struct vhost *v, struct client *c)
 	if (addr == NULL)
 		return 0;
 
-	if (*c->domain == '\0')
-		strlcpy(c->domain, addr->pp, sizeof(c->domain));
+	if (*c->domain == '\0') {
+		if (strlcpy(c->domain, addr->pp, sizeof(c->domain))
+		    >= sizeof(c->domain)) {
+			log_warnx("%s: domain too long: %s", __func__,
+			    addr->pp);
+			*c->domain = '\0';
+		}
+	}
 
 	if (matches(v->domain, c->domain))
 		return 1;
@@ -384,7 +390,7 @@ handle_handshake(int fd, short ev, void *d)
 		return;
 	default:
 		/* unreachable */
-		abort();
+		fatalx("unexpected return value from tls_handshake");
 	}
 
 	c->bev = bufferevent_new(fd, client_read, client_write,
@@ -608,8 +614,7 @@ apply_reverse_proxy(struct client *c)
 	if (p->reqca != NULL && check_matching_certificate(p->reqca, c))
 		return 1;
 
-	log_debug("opening proxy connection for %s:%s",
-	    p->host, p->port);
+	log_debug("opening proxy connection for %s:%s", p->host, p->port);
 
 	if ((c->pfd = proxy_socket(c, p->host, p->port)) == -1) {
 		start_reply(c, PROXY_ERROR, "proxy error");
@@ -713,8 +718,7 @@ apply_fastcgi(struct client *c)
 		return 0;
 	}
 
-	log_debug("opening fastcgi connection for (%s,%s)",
-	    f->path, f->port);
+	log_debug("opening fastcgi connection for (%s,%s)", f->path, f->port);
 
 	if (*f->port == '\0')
 		c->pfd = fcgi_open_sock(f);
@@ -813,8 +817,7 @@ open_dir(struct client *c)
 		return;
 	}
 
-	strlcpy(path, c->iri.path, sizeof(path));
-	strlcat(path, index, sizeof(path));
+	snprintf(path, sizeof(path), "%s%s", c->iri.path, index);
 
 	close(c->pfd);
 	c->pfd = fd;
@@ -952,6 +955,8 @@ client_read(struct bufferevent *bev, void *d)
 	struct evbuffer	*src = EVBUFFER_INPUT(bev);
 	const char	*path, *p, *parse_err = "invalid request";
 	char		 decoded[DOMAIN_NAME_LEN];
+	char		*nul;
+	size_t		 len;
 
 	bufferevent_disable(bev, EVBUFFER_READ);
 
@@ -982,6 +987,14 @@ client_read(struct bufferevent *bev, void *d)
 		return;
 	}
 
+	nul = strchr(c->req, '\0');
+	len = nul - c->req;
+	if (len != c->reqlen) {
+		log_debug("NUL inside the request IRI");
+		start_reply(c, BAD_REQUEST, "bad request");
+		return;
+	}
+
 	if (!parse_iri(c->req, &c->iri, &parse_err) ||
 	    !puny_decode(c->iri.host, decoded, sizeof(decoded), &parse_err)) {
 		log_debug("IRI parse error: %s", parse_err);
@@ -1295,7 +1308,8 @@ server_accept(int sock, short et, void *d)
 		if (errno == EWOULDBLOCK || errno == EAGAIN ||
 		    errno == ECONNABORTED)
 			return;
-		fatal("accept");
+		log_warnx("accept failed");
+		return;
 	}
 
 	mark_nonblock(fd);
